from datetime import datetime, timedelta, timezone

from asyncpg import Record

from udata_hydra import config, context


class Resource:
    """Represents a resource in the "catalog" DB table"""

    STATUSES = {
        None: "no status, waiting",
        "BACKOFF": "backoff period for this domain, will be checked later",
        "CRAWLING_URL": "resource URL currently being crawled",
        "TO_ANALYSE_RESOURCE": "resource to be processed for change, type and size analysis",
        "ANALYSING_RESOURCE": "currently being processed for change, type and size analysis",
        "TO_ANALYSE_CSV": "resource content to be analysed by CSV detective",
        "ANALYSING_CSV": "resource content currently being analysed by CSV detective",
        "INSERTING_IN_DB": "currently being inserted in DB",
        "CONVERTING_TO_PARQUET": "currently being converted to Parquet",
        "TO_ANALYSE_GEOJSON": "geojson resource content to be analysed",
        "ANALYSING_GEOJSON": "geojson resource content currently being analysed",
        "CONVERTING_TO_PMTILES": "currently being converted to pmtiles",
    }

    @classmethod
    async def get(cls, resource_id: str, column_name: str = "*") -> Record | None:
        pool = await context.pool()
        async with pool.acquire() as connection:
            q = f"""SELECT {column_name} FROM catalog WHERE resource_id = '{resource_id}';"""
            return await connection.fetchrow(q)

    @classmethod
    async def insert(
        cls,
        dataset_id: str,
        resource_id: str,
        url: str,
        type: str,
        format: str,
        status: str | None = None,
        priority: bool = True,
    ) -> None:
        if status and status not in cls.STATUSES.keys():
            raise ValueError(f"Invalid status: {status}")

        pool = await context.pool()
        async with pool.acquire() as connection:
            # Insert new resource in catalog table and mark as high priority for crawling
            q = """
                    INSERT INTO catalog (dataset_id, resource_id, url, type, format, deleted, status, priority)
                    VALUES ($1, $2, $3, $4, $5, FALSE, $6, $7)
                    ON CONFLICT (resource_id) DO UPDATE SET
                        dataset_id = $1,
                        url = $3,
                        type = $4,
                        format = $5,
                        deleted = FALSE,
                        status = $6,
                        priority = $7
                    RETURNING *;"""
            await connection.fetchrow(
                q, dataset_id, resource_id, url, type, format, status, priority
            )

    @classmethod
    async def update(cls, resource_id: str, data: dict) -> Record:
        """Update a resource in DB with new data and return the updated resource in DB"""
        columns = data.keys()
        # $1, $2...
        placeholders = [f"${x + 1}" for x in range(len(data.values()))]
        set_clause = ",".join([f"{c} = {v}" for c, v in zip(columns, placeholders)])
        pool = await context.pool()
        async with pool.acquire() as connection:
            q = f"""
                    UPDATE catalog
                    SET {set_clause}
                    WHERE resource_id = ${len(placeholders) + 1}
                    RETURNING *;"""
            return await connection.fetchrow(q, *data.values(), resource_id)

    @classmethod
    async def update_or_insert(
        cls,
        dataset_id: str,
        resource_id: str,
        url: str,
        type: str,
        format: str,
        status: str | None = None,
        priority: bool = True,  # Make resource high priority by default for crawling
    ) -> None:
        if status and status not in cls.STATUSES.keys():
            raise ValueError(f"Invalid status: {status}")

        pool = await context.pool()
        async with pool.acquire() as connection:
            # Check if resource is in catalog then insert or update into table
            if await Resource.get(resource_id):
                q = """
                        UPDATE catalog
                        SET dataset_id = $1, url = $3, type = $4, format=$5, status = $6, priority = $7
                        WHERE resource_id = $2
                        RETURNING *;"""
            else:
                q = """
                        INSERT INTO catalog (dataset_id, resource_id, url, type, format, deleted, status, priority)
                        VALUES ($1, $2, $3, $4, $5, FALSE, $6, $7)
                        ON CONFLICT (resource_id) DO UPDATE SET
                            dataset_id = $1,
                            url = $3,
                            type = $4,
                            format = $5,
                            deleted = FALSE,
                            status = $6,
                            priority = $7
                        RETURNING *;"""
            await connection.fetchrow(
                q, dataset_id, resource_id, url, type, format, status, priority
            )

    @classmethod
    async def delete(
        cls,
        resource_id: str,
    ) -> None:
        pool = await context.pool()
        async with pool.acquire() as connection:
            # Mark resource as deleted in catalog table
            q = f"""UPDATE catalog SET deleted = TRUE WHERE resource_id = '{resource_id}';"""
            await connection.execute(q)

    @staticmethod
    def get_excluded_clause() -> str:
        """Return the WHERE clause to get only resources from the check which:
        - don't have a URL in the excluded URLs patterns
        - are not deleted
        - are not currently being crawled or analysed (i.e. resources with no status, or status 'BACKOFF')
        """
        return " AND ".join(
            [f"catalog.url NOT LIKE '{p}'" for p in config.EXCLUDED_PATTERNS]
            + [
                "catalog.deleted = False",
                "(catalog.status IS NULL OR catalog.status = 'BACKOFF')",
            ]
        )

    @staticmethod
    async def get_stuck_resources() -> list[str]:
        """Some resources end up being stuck in a not null status forever,
        we want to get them back on track.
        This returns all resource ids of such stuck resources.
        """
        threshold = (
            datetime.now(timezone.utc) - timedelta(seconds=config.STUCK_THRESHOLD_SECONDS)
        ).strftime("%Y-%m-%d %H:%M:%S")
        q = f"""SELECT ca.resource_id
            FROM checks c
            JOIN catalog ca
            ON c.id = ca.last_check
            WHERE ca.status IS NOT NULL AND c.created_at < '{threshold}';"""
        pool = await context.pool()
        async with pool.acquire() as connection:
            rows = await connection.fetch(q)
        return [str(r["resource_id"]) for r in rows] if rows else []

    @classmethod
    async def clean_up_statuses(cls):
        stuck_resources: list[str] = await cls.get_stuck_resources()
        for rid in stuck_resources:
            await cls.update(rid, {"status": None})



import json

from asyncpg import Record

from udata_hydra import context


def convert_dict_values_to_json(data: dict) -> dict:
    """
    Convert values in dict that are dict to json for DB serialization
    TODO: this is suboptimal from asyncpg, dig into this
    https://magicstack.github.io/asyncpg/current/usage.html#example-automatic-json-conversion
    """
    for k, v in data.items():
        if type(v) is dict:
            data[k] = json.dumps(v)
    return data


def compute_insert_query(table_name: str, data: dict, returning: str = "id") -> str:
    columns = ",".join([f'"{k}"' for k in data.keys()])
    # $1, $2...
    placeholders = ",".join([f"${x + 1}" for x in range(len(data.values()))])
    return f"""
        INSERT INTO "{table_name}" ({columns})
        VALUES ({placeholders})
        RETURNING {returning};
    """


def compute_update_query(table_name: str, data: dict, returning: str = "*") -> str:
    columns = data.keys()
    # $1, $2...
    placeholders = [f"${x + 1}" for x in range(len(data.values()))]
    set_clause = ",".join([f"{c} = {v}" for c, v in zip(columns, placeholders)])
    return f"""
        UPDATE "{table_name}"
        SET {set_clause}
        WHERE id = ${len(placeholders) + 1}
        RETURNING {returning};
    """


async def update_table_record(table_name: str, record_id: int, data: dict) -> Record | None:
    data = convert_dict_values_to_json(data)
    q = compute_update_query(table_name, data)
    pool = await context.pool()
    return await pool.fetchrow(q, *data.values(), record_id)



import json
from datetime import date, datetime
from typing import IO

from asyncpg import Record
from dateparser import parse as date_parser
from dateutil.parser import ParserError
from dateutil.parser import parse as dateutil_parser

from udata_hydra import config
from udata_hydra.utils import UdataPayload, download_resource, queue, send


def to_json(value: str) -> str:
    """Convenience method, should be casted from string directly by postgres"""
    return value


def _parse_dt(value: str) -> datetime | None:
    """For performance reasons, we try first with dateutil and fallback on dateparser"""
    try:
        return dateutil_parser(value)
    except ParserError:
        return date_parser(value)


def to_date(value: str) -> date | None:
    parsed = _parse_dt(value)
    return parsed.date() if parsed else None


def to_datetime(value: str) -> datetime | None:
    return _parse_dt(value)


async def read_or_download_file(
    check: dict,
    file_path: str,
    file_format: str,
    exception: Record | None,
) -> IO[bytes]:
    return (
        open(file_path, "rb")
        if file_path
        else await download_resource(
            url=check["url"],
            headers=json.loads(check.get("headers") or "{}"),
            max_size_allowed=None
            if exception
            else int(config.MAX_FILESIZE_ALLOWED.get(file_format, "csv")),
        )
    )


async def notify_udata(resource: Record, check: dict) -> None:
    """Notify udata of the result of a parsing"""
    payload = {
        "resource_id": check["resource_id"],
        "dataset_id": resource["dataset_id"],
        "document": {
            "analysis:parsing:error": check["parsing_error"],
            "analysis:parsing:started_at": check["parsing_started_at"].isoformat()
            if check["parsing_started_at"]
            else None,
            "analysis:parsing:finished_at": check["parsing_finished_at"].isoformat()
            if check["parsing_finished_at"]
            else None,
        },
    }
    if config.CSV_TO_PARQUET and check.get("parquet_url"):
        payload["document"]["analysis:parsing:parquet_url"] = check.get("parquet_url")
        payload["document"]["analysis:parsing:parquet_size"] = check.get("parquet_size")
    if config.GEOJSON_TO_PMTILES and check.get("pmtiles_url"):
        payload["document"]["analysis:parsing:pmtiles_url"] = check.get("pmtiles_url")
        payload["document"]["analysis:parsing:pmtiles_size"] = check.get("pmtiles_size")
    payload["document"] = UdataPayload(payload["document"])
    queue.enqueue(send, _priority="high", **payload)
